#include <stdarg.h>
#include <stddef.h>
#include <inttypes.h>
#include <xen/xen.h>
#include <xen/callback.h>
#include <xen/grant_table.h>
#include <xen/version.h>

#include "shared.h"

struct regs_32 {
    /* pushed onto stack before calling into C code */
    uint32_t eax;
    uint32_t ebx;
    uint32_t ecx;
    uint32_t edx;
    uint32_t esi;
    uint32_t edi;
    uint32_t ebp;
    uint32_t trapno;
    /* trap / fault / int created */
    uint32_t error;
    uint32_t eip;
    uint32_t cs;
    uint32_t eflags;
    uint32_t esp;
    uint32_t ss;
};

/* 32bit defines */
#define longmode  0
#define regs      regs_32
#define emulate   emulate_32
#define fix_sel   fix_sel32
#define fix_desc  fix_desc32
#define ureg_t    uint32_t
#define sreg_t    int32_t
#define PRIxREG   PRIx32
#define rax       eax
#define rbx       ebx
#define rcx       ecx
#define rdx       edx
#define rsi       esi
#define rdi       edi
#define rbp       ebp
#define rsp       esp
#define rip       eip

#include "emu.h"

/* emu-data.c */
#define MAPS_MAX 256
extern page_aligned uint32_t maps_32[];
extern page_aligned uint64_t maps_pae[];
extern uint32_t maps_refcnt[MAPS_MAX];

extern struct descriptor_32 page_aligned xen_gdt[8192];
extern struct descriptor_32 page_aligned xen_idt[256];
extern struct tss_32 page_aligned xen_tss;
extern struct vcpu_guest_context xen_ctxt;
extern struct shared_info page_aligned shared_info;

extern struct xen_callback  xencb[8];
extern struct trap_info     xentr[256];

/* emu-main.c */
int is_pae(void);
int panic(char *message);
asmlinkage void do_boot(struct regs_32 regs);
asmlinkage void do_illegal_instruction(struct regs_32 regs);
asmlinkage void do_general_protection(struct regs_32 regs);
asmlinkage void do_page_fault(struct regs_32 regs);
asmlinkage void do_event_callback(struct regs_32 regs);
asmlinkage void do_guest_forward(struct regs_32 regs);
asmlinkage void do_debug_int3(struct regs_32 regs);

/* emu-mm.c */
void *map_32(uint32_t maddr);
void *map_pae(uint64_t maddr);
void *map_page(uint64_t maddr);
void free_page(void *ptr);
uint32_t *find_pte_32_lpt(uint32_t va);
uint32_t *find_pte_32_map(uint32_t va);
uint64_t *find_pte_pae_lpt(uint32_t va);
uint64_t *find_pte_pae_map(uint32_t va);
void pgtable_walk_pae(uint32_t va);
void update_emu_mappings(uint32_t cr3_mfn);

/* emu-hcall.c */
asmlinkage void do_hypercall(struct regs_32 regs);

/* emu-instr.c */
int emulate_32(struct regs_32 *regs);

/* printk.c */
int vscnprintf(char *buf, size_t size, const char *fmt, va_list args);
int snprintf(char * buf, size_t size, const char *fmt, ...);
int printk(const char *fmt, ...) __attribute__((format(printf, 1, 2)));

/* macros */
#define context_is_emu(regs)    (((regs)->cs & 0x03) == 0x00)
#define context_is_kernel(regs) (((regs)->cs & 0x03) == 0x01)
#define context_is_user(regs)   (((regs)->cs & 0x03) == 0x03)
